using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Additional namespaces
using robotino = rec.robotino.api2;
using System.Text;
#endregion

namespace RobotinoMRDS.Services.Robotino
{
    [Contract(Contract.Identifier)]
    [DisplayName("Robotino")]
    [Description("Provides services for the Robotino robot")]
    class RobotinoService : DsspServiceBase
    {
        /// <summary>
        /// 
        /// </summary>
        Robotino _robotino = new Robotino();

        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        RobotinoState _state = new RobotinoState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/Robotino", AllowMultipleInstances = false)]
        RobotinoOperations _mainPort = new RobotinoOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        // Internal Bumper port
        private BumperOperations _bumperPort = new BumperOperations();

        // Internal DistanceSensors port
        private DistanceSensorsOperations _distanceSensorsPort = new DistanceSensorsOperations();

        /// <summary>
        /// Service constructor
        /// </summary>
        public RobotinoService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            Activate(
                Arbiter.Interleave(
                new TeardownReceiverGroup(
                    Arbiter.Receive<DsspDefaultDrop>(false, _mainPort, DefaultDropHandler)
                    ),
                    new ExclusiveReceiverGroup(
                        Arbiter.ReceiveWithIterator<Replace>(true, _mainPort, ReplaceHandler),
                        Arbiter.ReceiveWithIterator<OmniDrive>(true, _mainPort, OmniDriveHandler),
                        Arbiter.ReceiveWithIterator<BumperInput>(true, _bumperPort, BumperInputHandler),
                        Arbiter.ReceiveWithIterator<DistanceSensorsInput>(true, _distanceSensorsPort, DistanceSensorsInputHandler)
                    ),
                    new ConcurrentReceiverGroup(
                        Arbiter.ReceiveWithIterator<Get>(true, _mainPort, GetHandler),
                        Arbiter.ReceiveWithIterator<Subscribe>(true, _mainPort, SubscribeHandler),
                        Arbiter.ReceiveWithIterator<SelectiveSubscribe>(true, _mainPort, SelectiveSubscribeHandler)
                    )
            ));

            AssignHandlers();

            _robotino.Connect("192.168.5.5");
            //_robotino.Connect("127.0.0.1:12080");

            base.Start();
        }

        #region Robotino Event Handlers

        void AssignHandlers()
        {
            _robotino.ConnectedEvent += new Robotino.ConnectedEventHandler(_robotino_Connected);
            _robotino.DisconnectedEvent += new Robotino.DisconnectedEventHandler(_robotino_Disconnected);
            _robotino.ErrorEvent += new Robotino.ErrorEventHandler(_robotino_Error);
            _robotino.BumperEvent += new Robotino.BumperEventHandler(_robotino_Bumper);
            _robotino.DistanceSensorsEvent += new Robotino.DistanceSensorsEventHandler(_robotino_DistanceSensors);
        }

        void _robotino_Error(Robotino sender, string error)
        {
            Console.WriteLine("Error: " + error);
            //throw new NotImplementedException();
        }

        void _robotino_Connected(Robotino sender)
        {
            Console.WriteLine("Connected to Robotino.");
            _state.Connected = true;
        }

        void _robotino_Disconnected(Robotino sender)
        {
            Console.WriteLine("Disconnected from Robotino.");
            _state.Connected = false;
        }

        void _robotino_DistanceSensors(Robotino sender, float[] distances)
        {
            DistanceSensors newValue = new DistanceSensors();
            newValue.Distances = distances;

            DistanceSensorsInput newDistanceSensorData = new DistanceSensorsInput();
            newDistanceSensorData.Body = newValue;

            // Send message to self on internal port
            _distanceSensorsPort.Post(newDistanceSensorData);
        }

        void _robotino_Bumper(Robotino sender, bool value)
        {
            Bumper newValues = new Bumper();
            newValues.Value = value;

            BumperInput newBumperData = new BumperInput();
            newBumperData.Body = newValues;

            // Send message to self on internal port
            _bumperPort.Post(newBumperData);
        }

        #endregion

        #region Internal port handlers

        /// <summary>
        /// Handles change in Bumper sensor values from Robotino and notifies respective subscribers
        /// </summary>
        /// <param name="bumperInput"></param>
        /// <returns></returns>
        public IEnumerator<ITask> BumperInputHandler(BumperInput bumperInput)
        {
            if (_state.Bumper.Value!= bumperInput.Body.Value)
            {
                _state.Bumper.Value = bumperInput.Body.Value;
                
                // notify subscribers
                submgr.Submit sub = new submgr.Submit(_state, DsspActions.ReplaceRequest, "Bumper");
                _submgrPort.Post(sub);

                bumperInput.ResponsePort.Post(DefaultInsertResponseType.Instance);
            }
            yield break;
        }

        /// <summary>
        /// Handles change in Distance Sensor values from Robotino and notifies respective subscribers
        /// </summary>
        /// <param name="distanceSensorInput"></param>
        /// <returns></returns>
        public IEnumerator<ITask> DistanceSensorsInputHandler(DistanceSensorsInput distanceSensorsInput)
        {
            _state.DistanceSensors.Distances = distanceSensorsInput.Body.Distances;

            // notify subscribers
            submgr.Submit sub = new submgr.Submit(_state, DsspActions.ReplaceRequest, "DistanceSensors");
            _submgrPort.Post(sub);

            distanceSensorsInput.ResponsePort.Post(DefaultInsertResponseType.Instance);
            yield break;
        }

        #endregion

        #region MainPort Handlers

        public IEnumerator<ITask> OmniDriveHandler(OmniDrive omniDrive)
        {
            _robotino.OmniDrive.setVelocity(
                omniDrive.Body.VelocityX,
                omniDrive.Body.VelocityZ,
                omniDrive.Body.Omega);

            omniDrive.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        public IEnumerator<ITask> RotateHandler(Rotate rotate)
        {
            throw new NotImplementedException("RotateHandler not implemented");
        }

        /// <summary>
        /// Handles Subscribe messages
        /// </summary>
        /// <param name="subscribe">the subscribe request</param>
        [ServiceHandler]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            base.SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }

        /// <summary>
        /// Handles CustomSubscribe messages
        /// </summary>
        /// <param name="subscribe">the subscribe request</param>
        [ServiceHandler]
        public IEnumerator<ITask> SelectiveSubscribeHandler(SelectiveSubscribe subscribe)
        {
            submgr.InsertSubscription insert = new submgr.InsertSubscription(
                new submgr.InsertSubscriptionMessage(
                    subscribe.Body.Subscriber,
                    subscribe.Body.Expiration,
                    0));

            insert.Body.NotificationCount = subscribe.Body.NotificationCount;

            List<submgr.QueryType> subscribeFilter = new List<submgr.QueryType>();
            foreach (string s in subscribe.Body.TypeFilter)
            {
                subscribeFilter.Add(new submgr.QueryType(s));
            }

            insert.Body.QueryList = subscribeFilter.ToArray();
            // These filters are interpreted as regular expressions
            //insert.Body.FilterType = submgr.FilterType.Regex;

            _submgrPort.Post(insert);

            yield return Arbiter.Choice(
                insert.ResponsePort,
                delegate(SubscribeResponseType rsp)
                {
                    subscribe.ResponsePort.Post(rsp);
                },
                delegate(Fault fault)
                {
                    subscribe.ResponsePort.Post(fault);
                }
            );
        }

        /// <summary>
        /// Handles Get messages
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Handles Replace messages
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }
        #endregion
    }

    #region Robotino Class
    public class Robotino
    {
        public delegate void ConnectedEventHandler(Robotino sender);
        public delegate void DisconnectedEventHandler(Robotino sender);
        public delegate void ErrorEventHandler(Robotino sender, String error);
        public delegate void BumperEventHandler(Robotino sender, bool value);
        public delegate void DistanceSensorsEventHandler(Robotino sender, float[] distances);

        public event ConnectedEventHandler ConnectedEvent;
        public event DisconnectedEventHandler DisconnectedEvent;
        public event ErrorEventHandler ErrorEvent;
        public event BumperEventHandler BumperEvent;
        public event DistanceSensorsEventHandler DistanceSensorsEvent;

        protected readonly robotino.Com _com;
        protected readonly robotino.OmniDrive _omniDrive;
        protected readonly robotino.Bumper _bumper;
        protected readonly robotino.DistanceSensorArray _distanceSensors;

        public Robotino()
        {
            _com = new RobotinoCom(this);
            _omniDrive = new robotino.OmniDrive();
            _bumper = new RobotinoBumper(this);
            _distanceSensors = new RobotinoDistanceSensors(this);

            _omniDrive.setComId(_com.id());
            _bumper.setComId(_com.id());
            _distanceSensors.setComId(_com.id());
        }

        public robotino.Com Com
        {
            get { return _com; }
        }

        public robotino.OmniDrive OmniDrive
        {
            get { return _omniDrive; }
        }

        public robotino.Bumper Bumper
        {
            get { return _bumper; }
        }

        public robotino.DistanceSensorArray DistanceSensors
        {
            get { return _distanceSensors; }
        }

        public virtual void Connect(String hostname)
        {
            _com.setAddress(hostname);
            _com.connectToServer();
        }

        private class RobotinoCom : robotino.Com
        {
            Robotino _robotino;
            System.Timers.Timer _spinTimer;

            public RobotinoCom(Robotino robotino)
            {
                this._robotino = robotino;

                _spinTimer = new System.Timers.Timer();
                _spinTimer.Elapsed += new System.Timers.ElapsedEventHandler(onSpinTimerTimeout);
                _spinTimer.Interval = 20;
                _spinTimer.Enabled = true;
            }

            public void onSpinTimerTimeout(object obj, System.Timers.ElapsedEventArgs e)
            {
                processEvents();
            }

            public override void connectedEvent()
            {
                //_robotino.isConnected = true;
                if (_robotino.ConnectedEvent != null)
                    _robotino.ConnectedEvent.BeginInvoke(_robotino, null, null);
            }

            public override void connectionClosedEvent()
            {
                //_robotino.isConnected = false;
                if (_robotino.DisconnectedEvent != null)
                    _robotino.DisconnectedEvent.BeginInvoke(_robotino, null, null);
            }

            public override void errorEvent(String errorStr)
            {
                if (_robotino.ErrorEvent != null)
                    _robotino.ErrorEvent.BeginInvoke(_robotino, errorStr, null, null);
            }
        }

        private class RobotinoBumper : robotino.Bumper
        {
            Robotino _robotino;
            private bool _bumped;

            // Default constructor
            public RobotinoBumper(Robotino robotino)
            {
                this._robotino = robotino;
            }

            public bool Bumped
            {
                get { return _bumped; }
                set { _bumped = value; }
            }

            public override void bumperEvent(bool hasContact)
            {
                _bumped = hasContact;

                if (_robotino.BumperEvent != null)
                    _robotino.BumperEvent.BeginInvoke(_robotino, _bumped, null, null);
            }
        }

        private class RobotinoDistanceSensors : robotino.DistanceSensorArray
        {
            Robotino _robotino;
            private float[] _distances = new float[] {0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f};
            
            // Default Constructor
            public RobotinoDistanceSensors(Robotino robotino)
            {
                this._robotino = robotino;
            }

            public float[] Distances
            {
                get { return _distances; }
                set { _distances = value; }
            }

            public override void distancesChangedEvent(float[] distances, uint size)
            {
                Array.Copy(distances, _distances, (int)size);

                if (_robotino.DistanceSensorsEvent != null)
                    _robotino.DistanceSensorsEvent.BeginInvoke(_robotino, _distances, null, null);
            }
        }
    }
    #endregion
}


