using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

namespace RobotinoMRDS.Services.Robotino
{
    /// <summary>
    /// Robotino contract class
    /// </summary>
    public sealed class Contract
    {
        /// <summary>
        /// DSS contract identifer for Robotino
        /// </summary>
        [DataMember]
        public const string Identifier = "http://servicerobotics.eu/2011/11/robotino.html";
    }

    /// <summary>
    /// Robotino main operations port
    /// </summary>
    [ServicePort]
    public class RobotinoOperations : PortSet<
        DsspDefaultLookup, 
        DsspDefaultDrop, 
        Get, 
        Subscribe,
        SelectiveSubscribe,
        Replace,
        OmniDrive,
        Rotate>
    {
    }

    #region Internal ports
    /// <summary>
    /// Internal port for Bumper updates
    /// </summary>
    internal class BumperOperations : Port<BumperInput>
    {
    }

    /// <summary>
    /// Internal port for DistanceSensors updates
    /// </summary>
    internal class DistanceSensorsOperations : Port<DistanceSensorsInput>
    {
    }
    #endregion

    #region Main port operations
    /// <summary>
    /// Robotino get operation
    /// </summary>
    public class Get : Get<GetRequestType, PortSet<RobotinoState, Fault>>
    {
    }

    /// <summary>
    /// Replace the current Robotino state
    /// </summary>
    public class Replace : Replace<RobotinoState, PortSet<DefaultReplaceResponseType, Fault>>
    {
    }

    /// <summary>
    /// Robotino Subscribe operation
    /// </summary>
    public class Subscribe : Subscribe<SubscribeRequestType, PortSet<SubscribeResponseType, Fault>>
    {
    }

    /// <summary>
    /// Robotino Custom Subscribe operation
    /// </summary>
    [Description("SelectiveSubscribe")]
    public class SelectiveSubscribe : Subscribe<SelectiveSubscribeRequestType, PortSet<SubscribeResponseType, Fault>>
    {
    }

    /// <summary>
    /// Robotino OmniDrive operation
    /// </summary>
    [Description("OmniDrive")]
    public class OmniDrive : Update<OmniDriveRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
    }

    /// <summary>
    /// Robotino Rotate operation
    /// </summary>
    [Description("Rotate")]
    public class Rotate : Update<RotateRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
    }

    #endregion

    #region Internal Sensor event operations
    /// <summary>
    /// Robotino Bumper Event operation
    /// </summary>
    [Description("BumperInput")]
    public class BumperInput : Insert<Bumper, PortSet<DefaultInsertResponseType, Fault>>
    {
    }

    /// <summary>
    /// Robotino Distance Sensors Event operation
    /// </summary>
    [Description("DistanceSensorsInput")]
    public class DistanceSensorsInput : Insert<DistanceSensors, PortSet<DefaultInsertResponseType, Fault>>
    {
    }
    #endregion

    /// <summary>
    /// For custom subscriptions
    /// </summary>
    [DataContract]
    public class SelectiveSubscribeRequestType : SubscribeRequestType
    {
        [Description("Identifies the names of the sensors")]
        public List<string> _sensorNames;

        [DataMember]
        public List<string> SensorNames
        {
            get { return _sensorNames; }
            set { _sensorNames = value; }
        }
    }

    /// <summary>
    /// OmniDrive request
    /// </summary>
    [DataMemberConstructor]
    [DataContract]
    public class OmniDriveRequestType
    {
        float _velocityX;
        float _velocityY;
        float _omega;

        [DataMember]
        public float VelocityX
        {
            get { return _velocityX; }
            set { _velocityX = value; }
        }

        [DataMember]
        public float VelocityZ
        {
            get { return _velocityY; }
            set { _velocityY = value; }
        }
        [DataMember]

        public float Omega
        {
            get { return _omega; }
            set { _omega = value; }
        }

        public OmniDriveRequestType()
        {
        }
    }

    /// <summary>
    /// Rotate request
    /// </summary>
    [DataMemberConstructor]
    [DataContract]
    public class RotateRequestType
    {
        int _direction;
        double _power;

        [DataMember]
        public int Direction
        {
            get { return _direction; }
            set { _direction = value; }
        }

        [DataMember]
        public double Power
        {
            get { return _power; }
            set { _power = value; }
        }

        public RotateRequestType()
        {
        }
    }
}


