using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Addition namespaces
using robotino = RobotinoMRDS.Services.Robotino.Proxy;
using contactsensor = Microsoft.Robotics.Services.ContactSensor.Proxy;
using http = Microsoft.Dss.Core.DsspHttp;
using System.Net;
#endregion

namespace RobotinoMRDS.Services.RobotinoServices.RobotinoBumperService
{
    [Contract(Contract.Identifier)]
    [DisplayName("RobotinoBumperService")]
    [Description("Provides access to Robotino's Bumper")]
    [AlternateContract(contactsensor.Contract.Identifier)]
    class RobotinoBumperService : DsspServiceBase
    {
        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        contactsensor.ContactSensorArrayState _state = new contactsensor.ContactSensorArrayState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/RobotinoBumperService", AllowMultipleInstances = true)]
        contactsensor.ContactSensorArrayOperations _mainPort = new contactsensor.ContactSensorArrayOperations();

        [Partner("Robotino",
            Contract = robotino.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate,
            Optional = false)]
        robotino.RobotinoOperations _robotinoPort = new robotino.RobotinoOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Service constructor
        /// </summary>
        public RobotinoBumperService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
                _state = new contactsensor.ContactSensorArrayState();

            _state.Sensors = new List<contactsensor.ContactSensor>();
            contactsensor.ContactSensor bumper = new contactsensor.ContactSensor();
            bumper.HardwareIdentifier = 0;
            bumper.Name = "Robotino Bumper";
            _state.Sensors.Add(bumper);

            base.Start();

            // Subscribe to Robotino
            SubscribeToRobotino();
        }

        private void SubscribeToRobotino()
        {
            // Create a notification port
            robotino.RobotinoOperations notificationPort = new robotino.RobotinoOperations();

            // Create a custom subscription request
            robotino.SelectiveSubscribeRequestType request = new robotino.SelectiveSubscribeRequestType();

            // Select the Bumper sensor only
            request.SensorNames = new List<string>();
            request.SensorNames.Add("Bumper");

            // Subscribe to Robotino and wait for a response
            Activate(
                Arbiter.Choice(_robotinoPort.SelectiveSubscribe(request, notificationPort),
                delegate(SubscribeResponseType rsp)
                {
                    // Subscription was successful, start listening for sensor change notifications
                    Activate(
                        Arbiter.Receive<robotino.Replace>(true, notificationPort, BumperNotificationHandler)
                        );
                },
                delegate(Fault f)
                {
                    LogError("Subscription failed: " + f.ToString());
                }
            )
            );

        }

        private void BumperNotificationHandler(robotino.Replace msg)
        {
            if (msg == null)
                throw new ArgumentNullException("Msg is null");

            // Update the state
            bool changed = (_state.Sensors[0].Pressed != msg.Body.Bumper.Value);
            _state.Sensors[0].TimeStamp = DateTime.Now;
            _state.Sensors[0].Pressed = msg.Body.Bumper.Value;

            if (changed)
            {
                // Notify subscribers if Bumper is pressed or not
                base.SendNotification<contactsensor.Replace>(_submgrPort, _state);
            }
        }

        #region ContactSensor handlers
        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(contactsensor.Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// HttpGet Handler
        /// </summary>
        /// <param name="httpGet"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> HttpGetHandler(http.HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(
                new http.HttpResponseType(
                    HttpStatusCode.OK,
                    _state
                )
            );
            yield break;
        }

        /// <summary>
        /// Replace Handler
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(contactsensor.Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// Update Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> UpdateHandler(contactsensor.Update update)
        {
            LogError("Bumper does not support inbound Update requests");
            throw new InvalidOperationException();
        }
        #endregion

        #region Subscribe handlers
        /// <summary>
        /// Handles Subscribe messages
        /// </summary>
        /// <param name="subscribe">the subscribe request</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(contactsensor.Subscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                            subscribe.Body.Subscriber, DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception fault)
                {
                    LogError(fault);
                }
            ));
            yield break;
        }

        /// <summary>
        /// Reliable Subscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(contactsensor.ReliableSubscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                            subscribe.Body.Subscriber, DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception fault)
                {
                    LogError(fault);
                }
            ));
            yield break;
        }
        #endregion
    }

    /// <summary>
    /// RobotinoBumperService contract class
    /// </summary>
    public sealed class Contract
    {
        /// <summary>
        /// DSS contract identifer for RobotinoBumperService
        /// </summary>
        [DataMember]
        public const string Identifier = "http://servicerobotics.eu/2011/11/robotinobumperservice.html";
    }
}


