﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Additional namespaces
using robotino = RobotinoMRDS.Services.Robotino.Proxy;
using analogsensorarray = Microsoft.Robotics.Services.AnalogSensorArray.Proxy;
using analogsensor = Microsoft.Robotics.Services.AnalogSensor.Proxy;
using http = Microsoft.Dss.Core.DsspHttp;
using System.Net;
#endregion

namespace RobotinoMRDS.Services.RobotinoServices.RobotinoDistanceSensorsService
{
    [Contract(Contract.Identifier)]
    [DisplayName("RobotinoDistanceSensorsService")]
    [Description("Provides access to Robotino's DistanceSensors")]
    [AlternateContract(analogsensorarray.Contract.Identifier)]
    class RobotinoDistanceSensorsService : DsspServiceBase
    {
        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        analogsensorarray.AnalogSensorArrayState _state = new analogsensorarray.AnalogSensorArrayState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/RobotinoDistanceSensorsService", AllowMultipleInstances = true)]
        analogsensorarray.AnalogSensorOperations _mainPort = new analogsensorarray.AnalogSensorOperations();

        [Partner("Robotino",
            Contract = robotino.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate,
            Optional = false)]
        robotino.RobotinoOperations _robotinoPort = new robotino.RobotinoOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Empty constructor
        /// </summary>
        public RobotinoDistanceSensorsService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
                _state = new analogsensorarray.AnalogSensorArrayState();

            _state.Sensors = new List<analogsensor.AnalogSensorState>();

            for (int i = 0; i < 9; i++)
            {
                analogsensor.AnalogSensorState distanceSensor = new analogsensor.AnalogSensorState();
                distanceSensor.HardwareIdentifier = i;
                _state.Sensors.Add(distanceSensor);

            }
            base.Start();

            // Subscribe to Robotino
            SubscribeToRobotino();
        }

        private void SubscribeToRobotino()
        {
            // Create a notification port
            robotino.RobotinoOperations notificationPort = new robotino.RobotinoOperations();

            // Create a custom subscription request
            robotino.SelectiveSubscribeRequestType request = new robotino.SelectiveSubscribeRequestType();

            // Select the DistanceSensors only
            request.SensorNames = new List<string>();
            request.SensorNames.Add("DistanceSensors");

            // Subscribe to Robotino and wait for a response
            Activate(
                Arbiter.Choice(_robotinoPort.SelectiveSubscribe(request, notificationPort),
                delegate(SubscribeResponseType rsp)
                {
                    // Subscription was successful, start listening for sensor change notifications
                    Activate(
                        Arbiter.Receive<robotino.Replace>(true, notificationPort, DistanceSensorsNotificationHandler)
                        );
                },
                delegate(Fault f)
                {
                    LogError("Subscription failed: " + f.ToString());
                }
            )
            );
        }

        private void DistanceSensorsNotificationHandler(robotino.Replace msg)
        {
            if (msg == null)
                throw new ArgumentNullException("Msg is null");
            //bool changed = false;

            foreach (analogsensor.AnalogSensorState distanceSensor in _state.Sensors)
            {
                distanceSensor.TimeStamp = DateTime.Now;
                distanceSensor.RawMeasurement = msg.Body.DistanceSensors.Distances[distanceSensor.HardwareIdentifier];
            }

            base.SendNotification<analogsensorarray.Replace>(_submgrPort, _state);
        }

        #region AnalogSensorArray handlers
        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(analogsensorarray.Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// HttpGet Handler
        /// </summary>
        /// <param name="httpGet"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> HttpGetHandler(http.HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(
                new http.HttpResponseType(
                    HttpStatusCode.OK,
                    _state
                )
            );
            yield break;
        }

        /// <summary>
        /// Replace Handler
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(analogsensorarray.Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }
        #endregion

        #region Subscribe handlers
        /// <summary>
        /// Handles Subscribe messages
        /// </summary>
        /// <param name="subscribe">the subscribe request</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(analogsensorarray.Subscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                            subscribe.Body.Subscriber, DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception ex)
                {
                    LogError(ex);
                }
            ));
            yield break;
        }

        /// <summary>
        /// Reliable Subscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(analogsensorarray.ReliableSubscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                            subscribe.Body.Subscriber, DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception ex)
                {
                    LogError(ex);
                }
            ));
            yield break;
        }
        #endregion
    }

    /// <summary>
    /// RobotinoDistanceSensorsService contract class
    /// </summary>
    public sealed class Contract
    {
        /// <summary>
        /// DSS contract identifer for RobotinoDistanceSensorsService
        /// </summary>
        [DataMember]
        public const string Identifier = "http://servicerobotics.eu/2012/03/robotinodistancesensorsservice.html";
    }
}
