﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

#region Simulation namespaces
using Microsoft.Robotics.Simulation;
using Microsoft.Robotics.Simulation.Engine;
using Microsoft.Robotics.Simulation.Physics;
using Microsoft.Robotics.PhysicalModel;
using xna = Microsoft.Xna.Framework;
using xnagrfx = Microsoft.Xna.Framework.Graphics;
#endregion

namespace RobotinoMRDS.Simulation.Robotino
{
    /// <summary>
    /// Models an array of contact sensors
    /// </summary>
    [DataContract]
    public class RobotinoBumperArrayEntity : VisualEntity
    {
        Port<EntityContactNotification> _notifications = new Port<EntityContactNotification>();

        BoxShape[] _shapes;

        /// <summary>
        /// Shapes for each bumper
        /// </summary>
        [DataMember]
        [Description("Contains the shapes which represent the bumpers.")]
        public BoxShape[] Shapes
        {
            get { return _shapes; }
            set { _shapes = value; }
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public RobotinoBumperArrayEntity()
        {
        }

        /// Initialization constructor used when this entity is built programmatically
        /// </summary>
        /// <param name="initialPos"></param>
        public RobotinoBumperArrayEntity(string name, Pose pose)
        {
            base.State.Name = name;
            base.State.Pose = pose;
        }

        /// <summary>
        /// Initialization
        /// </summary>
        /// <param name="device"></param>
        /// <param name="physicsEngine"></param>
        public override void Initialize(xnagrfx.GraphicsDevice device, PhysicsEngine physicsEngine)
        {
            try
            {
                InitError = string.Empty;

                // Shapes
                string[] bumperName = new string[] { "B1", "B2", "B3", "B4", "B5", 
                "B6", "B7", "B8", "B9", "B10",
                "B11", "B12", "B13", "B14", "B15",
                "B16", "B17", "B18", "B19", "B20"};

                float start = (float)(-Math.PI);
                float end = (float)(Math.PI);
                int segments = 20; // 20 bumpers
                float increment = (end - start) / (segments - 1);
                int segment = 0;

                _shapes = new BoxShape[segments];

                for (float angle = start; angle <= end; angle += increment, segment++)
                {
                    BoxShape bumper = new BoxShape(
                        new BoxShapeProperties(
                            bumperName[segment], 0.001f, //mass
                            new Pose(new Vector3(0.16f * (float)(Math.Sin(angle)), 0.03f,
                                0.16f * (float)(Math.Cos(angle))),
                                Quaternion.FromAxisAngle(0, 1, 0, angle)), //rotation
                                new Vector3(0.05f, 0.15f, 0.01f))); //dimensions

                    bumper.BoxState.Material = new MaterialProperties("Bumper", 0.0f, 0.25f, 0.5f);
                    bumper.Parent = this;
                    bumper.BoxState.EnableContactNotifications = true;

                    _shapes[segment] = bumper;
                }

                foreach (BoxShape b in _shapes)
                {
                    State.PhysicsPrimitives.Add(b);
                }

                base.Initialize(device, physicsEngine);

                if (Parent == null)
                    throw new Exception("This entity must be a child of another entity.");


                CreateAndInsertPhysicsEntity(physicsEngine);

                Flags |= VisualEntityProperties.DisableRendering;
            }
            catch (Exception ex)
            {
                HasBeenInitialized = false;
                InitError = ex.ToString();
            }
        }

        /// <summary>
        /// Adds a notification port to the list of subscriptions that get notified when the bumper shapes
        /// collide in the physics world
        /// </summary>
        /// <param name="notificationTarget"></param>
        public void Subscribe(Port<EntityContactNotification> notificationTarget)
        {
            // the parent has the physics entity create (we are just part of that entity)
            // so subscribe there.
            PhysicsEntity.SubscribeForContacts(notificationTarget);
        }
    }
}
