﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

#region Simulation namespaces
using Microsoft.Robotics.Simulation;
using Microsoft.Robotics.Simulation.Engine;
using engineproxy = Microsoft.Robotics.Simulation.Engine.Proxy;
using Microsoft.Robotics.Simulation.Physics;
using Microsoft.Robotics.PhysicalModel;
using xna = Microsoft.Xna.Framework;
#endregion

namespace RobotinoMRDS.Simulation.Robotino
{
    [DataContract]
    [DataMemberConstructor]
    public class RobotinoEntity : VisualEntity
    {
        #region Instance variables

        // Robotino's wheels
        RobotinoWheel0Entity _wheel0;
        RobotinoWheel1Entity _wheel1;
        RobotinoWheel2Entity _wheel2;

        //True if drive mechanism is enabled
        bool _isEnabled;

        //Scaling factor to apply to motor torque requests
        float _motorTorqueScaling;

        float _wheel0TargetVelocity;
        float _wheel1TargetVelocity;
        float _wheel2TargetVelocity;


        const float MAX_VELOCITY = 20.0f;
        const float MIN_VELOCITY = -MAX_VELOCITY;

        // OmniDrive variables
        SuccessFailurePort _omniDrivePort = null;
        bool _omniDriveActive = false;

        // DriveDistance variables
        Pose _startPoseForDriveDistance;
        double _distanceToTravel;
        SuccessFailurePort _driveDistancePort = null;

        //RotateDegrees variables
        double _targetRotation = double.MaxValue;
        double _currentRotation = 0;
        double _previousHeading = 0;
        const float acceptableRotationError = 0.005f;
        SuccessFailurePort _rotateDegreesPort = null;

        // DriveDistance and OmniDrive variables
        Queue<double> progressPoints = new Queue<double>();
        const int averageKernel = 6;
        const int decelerateThreshold = 6;
        const float twoPI = (float)(2 * Math.PI);

        //Rotate variables
        bool _rotateActive = false;
        SuccessFailurePort _rotatePort = null;

        const float SPEED_DELTA = 0.25f;
        #endregion

        [Category("Wheels")]
        [DataMember]
        public RobotinoWheel0Entity Wheel0
        {
            get { return _wheel0; }
            set { _wheel0 = value; }
        }

        [Category("Wheels")]
        [DataMember]
        public RobotinoWheel1Entity Wheel1
        {
            get { return _wheel1; }
            set { _wheel1 = value; }
        }

        [Category("Wheels")]
        [DataMember]
        public RobotinoWheel2Entity Wheel2
        {
            get { return _wheel2; }
            set { _wheel2 = value; }
        }

        [Category("Wheels")]
        [DataMember]
        public int Wheel0Encoder
        {
            get { return (int)(_wheel0.Rotations * 600f); }
            set { _wheel0.Rotations = (float)value / 600f; }
        }

        [Category("Wheels")]
        [DataMember]
        public int Wheel1Encoder
        {
            get { return (int)(_wheel1.Rotations * 600f); }
            set { _wheel1.Rotations = (float)value / 600f; }
        }

        [Category("Wheels")]
        [DataMember]
        public int Wheel2Encoder
        {
            get { return (int)(_wheel2.Rotations * 600f); }
            set { _wheel2.Rotations = (float)value / 600f; }
        }

        [DataMember]
        public bool IsEnabled
        {
            get { return _isEnabled; }
            set { _isEnabled = value; }
        }

        [DataMember]
        public float MotorTorqueScaling
        {
            get { return _motorTorqueScaling; }
            set { _motorTorqueScaling = value; }
        }


        //Default constructor
        public RobotinoEntity()
        {
        }

        //Initialization constructor used when this entity is built programmatically
        public RobotinoEntity(string name, Pose initialPose)
        {
            base.State.Name = name;
            base.State.Pose = initialPose;
            _motorTorqueScaling = 20.0f;

            // The Camera entity
            CameraEntity frontCam = new CameraEntity(
                320,
                240,
                (float)(30.0 * Math.PI / 180));
            frontCam.State.Name = name + "_cam";
            frontCam.IsRealTimeCamera = true; //update each frame
            frontCam.State.Pose.Position = new Vector3(
                0,
                chassisClearance + 0.17f,
                -5.5f / 100);
            InsertEntityGlobal(frontCam);

            // The LRF entity
            LaserRangeFinderEntity lrf = new LaserRangeFinderEntity(
                new Pose(new Vector3(
                    0,
                    chassisClearance + 0.28f,
                    2.0f / 100)));
            lrf.State.Name = name + "_lrf";
            lrf.State.Assets.Effect = "LaserRangeFinder.fx";
            InsertEntityGlobal(lrf);

            // The Bumper Array Entity
            RobotinoBumperArrayEntity bumper = new RobotinoBumperArrayEntity(
                base.State.Name + "_bumper",
                new Pose(new Vector3(
                    0,
                    chassisClearance + 0.05f,
                    0)));
            bumper.Parent = this;
            InsertEntityGlobal(bumper);

            // The IR Array entity
            RobotinoIRArrayEntity irArray = new RobotinoIRArrayEntity(
                base.State.Name + "_irarray",
                new Pose(new Vector3(
                    0,
                    chassisClearance + 0.02f,
                    0)));
            irArray.Parent = this;
            InsertEntityGlobal(irArray);
        }

        public float CurrentHeading
        {
            get
            {
                //return the axis angle of the quaternion
                xna.Vector3 euler = UIMath.QuaternionToEuler(State.Pose.Orientation);
                //heading is the rotation about the Y Axis
                return xna.MathHelper.ToRadians(euler.Y);
            }
        }

        public void SetMotorTorque(float wheel0, float wheel1, float wheel2)
        {
            ResetRotationAndDistance();
            SetAxleVelocity(
                wheel0 * _motorTorqueScaling,
                wheel1 * _motorTorqueScaling,
                wheel2 * _motorTorqueScaling);
        }

        public void SetVelocity(float value)
        {
            ResetRotationAndDistance();
            SetVelocity(value, value, value);
        }

        //Sets angular velocity on the wheels
        public void SetVelocity(float wheel0, float wheel1, float wheel2)
        {
            ResetRotationAndDistance();
            if (_wheel0 == null || _wheel2 == null || _wheel1 == null)
                return;

            wheel0 = ValidateWheelVelocity(wheel0);
            wheel1 = ValidateWheelVelocity(wheel1);
            wheel2 = ValidateWheelVelocity(wheel2);
            //v is in m/sec - convert to an axle speed
            // 2Pi(V/2PiR) = V/R
            SetAxleVelocity(
                wheel0 / _wheel0.Wheel.State.Radius,
                wheel1 / _wheel1.Wheel.State.Radius,
                wheel2 / _wheel2.Wheel.State.Radius);
        }
        private void SetAxleVelocity(float wheel0, float wheel1, float wheel2)
        {
            _wheel0TargetVelocity = wheel0;
            _wheel1TargetVelocity = wheel1;
            _wheel2TargetVelocity = wheel2;
        }

        float ValidateWheelVelocity(float value)
        {
            if (value > MAX_VELOCITY)
                return MAX_VELOCITY;
            if (value < MIN_VELOCITY)
                return MIN_VELOCITY;

            return value;
        }

        /// <summary>
        /// Applies constant power to both wheels, driving the motor base for a fixed distance, in the current direction
        /// </summary>
        /// <param name="distance">Distance to travel, in meters</param>
        /// <param name="power">Normalized power (torque) value for both wheels</param>
        /// <param name="responsePort">A port to report the result of the request, success or exception</param>
        public void DriveDistance(
            float distance,
            float power,
            SuccessFailurePort responsePort)
        {
            ResetRotationAndDistance();

            //keep track of the reponse port for when we complete the request
            _driveDistancePort = responsePort;

            //handle -ve distances
            if (distance < 0)
            {
                distance = -distance;
                power = -power;
            }
            _startPoseForDriveDistance = State.Pose;
            _distanceToTravel = distance;
            SetAxleVelocity(
                power * _motorTorqueScaling,
                0.0f, //Should be 0.0f for wheel1 as it is useless here.
                -power * _motorTorqueScaling);
        }

        /// <summary>
        /// Calculates and sets motor velocities based on the input parameters.
        /// </summary>
        /// <param name="vx">The velocity along the X axis</param>
        /// <param name="vz">The velocity along the Z axis</param>
        /// <param name="responsePort">A port to report the result of the request, success or exception</param>
        public void OmniDrive(
            float vx,
            float vz,
            float omega,
            SuccessFailurePort responsePort)
        {
            ResetRotationAndDistance();

            _omniDriveActive = true;

            _omniDrivePort = responsePort;

            //Projection matrix
            float[] v0 = new float[2] { -0.5f * (float)Math.Sqrt(3.0), 0.5f };
            float[] v1 = new float[2] { 0.0f, 1.0f };
            float[] v2 = new float[2] { 0.5f * (float)Math.Sqrt(3.0), 0.5f };

            //Scale omega with the radius of the robot
            //127.0 here is the distance from the center to wheel center in mm
            float vOmegaScaled = 2.0f * (float)Math.PI * 127.0f * omega / 360.0f;

            //Convert from mm/s to RPM
            //16.0 is the gear value
            //37.5 is the radius of the wheel in mm
            const float k = 60.0f * 16.0f / (2.0f * (float)Math.PI * 37.5f);

            //Compute desired velocity
            float m0 = (v0[0] * vx + v0[1] * vz + vOmegaScaled) * k;
            float m1 = (v1[0] * vx + v1[1] * vz + vOmegaScaled) * k;
            float m2 = (v2[0] * vx + v2[1] * vz + vOmegaScaled) * k;

            m0 = ValidateWheelVelocity(m0);
            m1 = ValidateWheelVelocity(m1);
            m2 = ValidateWheelVelocity(m2);

            SetAxleVelocity(m0, m1, m2);
        }

        /// <summary>
        /// Applies constant power to each wheel (but of inverse polarity), rotating Robotino 
        /// through the given rotation.
        /// </summary>
        /// <param name="degrees">Rotation around Y axis, in degrees.</param>
        /// <param name="power">Normalized power (torque) value for both wheels</param>
        /// <param name="responsePort">A port to report the result of the request, success or exception</param>
        public void RotateDegrees(
            float degrees,
            float power,
            SuccessFailurePort responsePort)
        {
            ResetRotationAndDistance();

            _rotateDegreesPort = responsePort;

            _targetRotation = xna.MathHelper.ToRadians(degrees);
            _currentRotation = 0;
            _previousHeading = CurrentHeading;

            if (degrees < 0)
                SetAxleVelocity(
                    power * _motorTorqueScaling,
                    -power * _motorTorqueScaling,
                    power * _motorTorqueScaling);
            else
                SetAxleVelocity(
                    -power * _motorTorqueScaling,
                    power * _motorTorqueScaling,
                    -power * _motorTorqueScaling);
        }

        /// <summary>
        /// Applies constant power to each wheel (but of inverse polarity), rotating Robotino 
        /// continuosly
        /// </summary>
        /// <param name="direction">Direction of rotation (1 for CCW, 0 for CW)</param>
        /// <param name="power">Normalized power (torque) value for both wheels</param>
        /// <param name="responsePort">A port to report the result of the request, success or exception</param>
        public void Rotate(
            int direction,
            float power,
            SuccessFailurePort responsePort)
        {
            ResetRotationAndDistance();

            _rotatePort = responsePort;
            _rotateActive = true;

            if (direction == 1) //CCW
                SetAxleVelocity(
                    power * _motorTorqueScaling,
                    -power * _motorTorqueScaling,
                    power * _motorTorqueScaling);
            else //CW
                SetAxleVelocity(
                    -power * _motorTorqueScaling,
                    power * _motorTorqueScaling,
                    -power * _motorTorqueScaling);
        }

        /// <summary>
        /// When a direct update to motor torque or wheel velocity occurs
        /// we abandon any current DriveDistance, OmniDrive, Rotate or 
        /// RotateDegrees commands
        /// </summary>
        void ResetRotationAndDistance()
        {
            progressPoints.Clear();
            _distanceToTravel = 0;
            _targetRotation = double.MaxValue;
            _omniDriveActive = false;
            _rotateActive = false;

            if (_driveDistancePort != null)
            {
                _driveDistancePort.Post(
                    new Exception("Request superceded prior to completion."));
                _driveDistancePort = null;
            }
            if (_rotateDegreesPort != null)
            {
                _rotateDegreesPort.Post(
                    new Exception("Request superceded prior to completion."));
                _rotateDegreesPort = null;
            }
            if (_omniDrivePort != null)
            {
                _omniDrivePort.Post(
                    new Exception("Request superceded prior to completion."));
                _omniDrivePort = null;
            }
            if (_rotatePort != null)
            {
                _rotatePort.Post(
                    new Exception("Request superceded prior to completion."));
                _rotatePort = null;
            }
        }
        /// <summary>
        /// Updates the Robotino's state
        /// </summary>
        /// <param name="update"></param>
        public override void Update(FrameUpdate update)
        {
            PhysicsEntity.UpdateState(true);

            if (_distanceToTravel > 0)
            {
                //Drive distance update
                double currentDistance = Vector3.Length(State.Pose.Position -
                    _startPoseForDriveDistance.Position);
                if (currentDistance >= _distanceToTravel)
                {
                    _wheel0.Wheel.AxleSpeed = 0;
                    _wheel1.Wheel.AxleSpeed = 0;
                    _wheel2.Wheel.AxleSpeed = 0;

                    _wheel0TargetVelocity = 0;
                    _wheel1TargetVelocity = 0;
                    _wheel2TargetVelocity = 0;

                    _distanceToTravel = 0;
                    //now as we're finished, post a response
                    if (_driveDistancePort != null)
                    {
                        SuccessFailurePort tmp = _driveDistancePort;
                        _driveDistancePort = null;
                        tmp.Post(new SuccessResult());
                    }
                }
                else
                {
                    //need to drive further, check if we should slow down
                    if (progressPoints.Count >= averageKernel)
                    {
                        double distanceRemaining =
                            _distanceToTravel - currentDistance;
                        double framesToCompletion =
                            distanceRemaining * averageKernel /
                            (currentDistance - progressPoints.Dequeue());
                        if (framesToCompletion < decelerateThreshold)
                        {
                            _wheel0TargetVelocity *= 0.5f;
                            _wheel1TargetVelocity *= 0.05f;
                            _wheel2TargetVelocity *= 0.5f;
                            progressPoints.Clear();
                        }
                    }
                    progressPoints.Enqueue(currentDistance);
                }
                float left = _wheel0.Wheel.AxleSpeed + _wheel0TargetVelocity;
                float right = _wheel2.Wheel.AxleSpeed + _wheel2TargetVelocity;
                float rear = _wheel1.Wheel.AxleSpeed + _wheel1TargetVelocity;

                if (Math.Abs(left) > 0.1)
                {
                    if (left > 0)
                        _wheel0.Wheel.AxleSpeed -= SPEED_DELTA;
                    else
                        _wheel0.Wheel.AxleSpeed += SPEED_DELTA;
                }

                if (Math.Abs(right) > 0.1)
                {
                    if (right > 0)
                        _wheel2.Wheel.AxleSpeed -= SPEED_DELTA;
                    else
                        _wheel2.Wheel.AxleSpeed += SPEED_DELTA;
                }
            }

            else if (_targetRotation != double.MaxValue)
            {
                //RotateDegrees update
                float currentHeading = CurrentHeading;
                double angleDelta = currentHeading - _previousHeading;
                while (angleDelta > Math.PI)
                    angleDelta -= twoPI;
                while (angleDelta <= -Math.PI)
                    angleDelta += twoPI;
                _currentRotation += angleDelta;
                _previousHeading = currentHeading; //for next frame

                float angleError;

                if (_targetRotation < 0)
                    angleError = (float)(_currentRotation - _targetRotation);
                else
                    angleError = (float)(_targetRotation - _currentRotation);

                if (angleError < acceptableRotationError)
                {
                    // current heading is within acceptableError or has overshot
                    // end the rotation
                    _targetRotation = double.MaxValue;
                    _wheel0.Wheel.AxleSpeed = 0;
                    _wheel1.Wheel.AxleSpeed = 0;
                    _wheel2.Wheel.AxleSpeed = 0;

                    _wheel0TargetVelocity = 0;
                    _wheel1TargetVelocity = 0;
                    _wheel2TargetVelocity = 0;

                    // now that we're finished, post a response
                    if (_rotateDegreesPort != null)
                    {
                        SuccessFailurePort tmp = _rotateDegreesPort;
                        _rotateDegreesPort = null;
                        tmp.Post(new SuccessResult());
                    }
                }
                else
                {
                    if (angleDelta != 0)
                    {
                        // need to turn more, check to see if we should slow down
                        if (progressPoints.Count >= averageKernel)
                        {
                            double framesToCompletion = Math.Abs(angleError * averageKernel / (_currentRotation - progressPoints.Dequeue()));
                            if (framesToCompletion < decelerateThreshold)
                            {
                                if (Math.Abs(_wheel0TargetVelocity) > 0.1)
                                    _wheel0TargetVelocity *= 0.5f;

                                if (Math.Abs(_wheel1TargetVelocity) > 0.1)
                                    _wheel1TargetVelocity *= 0.5f;

                                if (Math.Abs(_wheel2TargetVelocity) > 0.1)
                                    _wheel2TargetVelocity *= 0.5f;

                                progressPoints.Clear();
                            }
                        }
                        progressPoints.Enqueue(_currentRotation);
                    }
                }
                float left = _wheel0.Wheel.AxleSpeed + _wheel0TargetVelocity;
                float right = _wheel2.Wheel.AxleSpeed + _wheel2TargetVelocity;
                float rear = _wheel1.Wheel.AxleSpeed + _wheel1TargetVelocity;

                if (Math.Abs(left) > 0.1)
                {
                    if (left > 0)
                        _wheel0.Wheel.AxleSpeed -= SPEED_DELTA;
                    else
                        _wheel0.Wheel.AxleSpeed += SPEED_DELTA;
                }

                if (Math.Abs(rear) > 0.1)
                {
                    if (rear > 0)
                        _wheel1.Wheel.AxleSpeed -= SPEED_DELTA;
                    else
                        _wheel1.Wheel.AxleSpeed += SPEED_DELTA;
                }

                if (Math.Abs(right) > 0.1)
                {
                    if (right > 0)
                        _wheel2.Wheel.AxleSpeed -= SPEED_DELTA;
                    else
                        _wheel2.Wheel.AxleSpeed += SPEED_DELTA;
                }
            }
            else if (_omniDriveActive == true)
            {
                if (_omniDrivePort != null)
                {
                    SuccessFailurePort tmpdrive = _omniDrivePort;
                    _omniDrivePort = null;
                    tmpdrive.Post(new SuccessResult());
                }

                _wheel0.Wheel.AxleSpeed = _wheel0TargetVelocity;
                _wheel1.Wheel.AxleSpeed = _wheel1TargetVelocity;
                _wheel2.Wheel.AxleSpeed = _wheel2TargetVelocity;
            }
            else if (_rotateActive == true)
            {
                if (_rotatePort != null)
                {
                    SuccessFailurePort tmprotate = _rotatePort;
                    _rotatePort = null;
                    tmprotate.Post(new SuccessResult());
                }

                _wheel0.Wheel.AxleSpeed = _wheel0TargetVelocity;
                _wheel1.Wheel.AxleSpeed = _wheel1TargetVelocity;
                _wheel2.Wheel.AxleSpeed = _wheel2TargetVelocity;
            }

            _wheel0.Update(update);
            _wheel1.Update(update);
            _wheel2.Update(update);

            base.Update(update);
        }

        static float mass = 11.0f; //kg
        static float chassisClearance = (1.4f / 100);

        static float wheelDiameter = (7.5f / 100);
        static float wheelMass = 0.07f; //kg

        static Vector3 chassisDimensions = new Vector3(
            15.4f / 100,
            7.0f / 100,
            15.4f / 100);

        static Vector3 wheel0Position = new Vector3(
            -11.0f / 100,
            wheelDiameter / 2.0f,
            -6.35f / 100);

        static Vector3 wheel1Position = new Vector3(
            0.0f,
            wheelDiameter / 2.0f,
            12.7f / 100);

        static Vector3 wheel2Position = new Vector3(
            -wheel0Position.X,
            wheelDiameter / 2.0f,
            -6.35f / 100);

        public override void Initialize(Microsoft.Xna.Framework.Graphics.GraphicsDevice device, PhysicsEngine physicsEngine)
        {
            try
            {
                //Chassis
                BoxShapeProperties chassisDesc = new BoxShapeProperties(
                    "chassis",
                    mass / 2.0f,
                    new Pose(new Vector3(
                        0,
                        chassisClearance + chassisDimensions.Y / 2.0f,
                        0)),
                        chassisDimensions);
                chassisDesc.Material = new MaterialProperties("chassisMaterial", 0.0f, 0.5f, 0.5f);

                BoxShape chassis = new BoxShape(chassisDesc);
                chassis.State.Name = "ChassisShape";
                State.PhysicsPrimitives.Add(chassis);

                CreateAndInsertPhysicsEntity(physicsEngine);
                PhysicsEntity.SolverIterationCount = 128;

                //Wheels
                WheelShapeProperties wheel0Prop = new WheelShapeProperties(
                    "WheelLeft",
                    wheelMass,
                    wheelDiameter / 2.0f);

                WheelShapeProperties wheel1Prop = new WheelShapeProperties(
                    "WheelRear",
                    wheelMass,
                    wheelDiameter / 2.0f);

                WheelShapeProperties wheel2Prop = new WheelShapeProperties(
                    "WheelRight",
                    wheelMass,
                    wheelDiameter / 2.0f);

                wheel0Prop.Flags |= WheelShapeBehavior.OverrideAxleSpeed;
                wheel1Prop.Flags |= WheelShapeBehavior.OverrideAxleSpeed;
                wheel2Prop.Flags |= WheelShapeBehavior.OverrideAxleSpeed;

                wheel0Prop.InnerRadius = 0.7f * wheelDiameter / 2.0f;
                wheel1Prop.InnerRadius = 0.7f * wheelDiameter / 2.0f;
                wheel2Prop.InnerRadius = 0.7f * wheelDiameter / 2.0f;

                wheel0Prop.LocalPose = new Pose(wheel0Position);
                wheel1Prop.LocalPose = new Pose(wheel1Position);
                wheel2Prop.LocalPose = new Pose(wheel2Position);

                wheel0Prop.LocalPose.Orientation = TypeConversion.FromXNA(xna.Quaternion.CreateFromAxisAngle(new xna.Vector3(0, 1, 0), -(float)(Math.PI / 6)));
                wheel1Prop.LocalPose.Orientation = TypeConversion.FromXNA(xna.Quaternion.CreateFromAxisAngle(new xna.Vector3(0, 1, 0), -(float)(Math.PI / 2)));
                wheel2Prop.LocalPose.Orientation = TypeConversion.FromXNA(xna.Quaternion.CreateFromAxisAngle(new xna.Vector3(0, 1, 0), (float)(Math.PI / 6 + Math.PI)));

                TireForceFunctionDescription LongitudalFunction = new TireForceFunctionDescription();
                LongitudalFunction.ExtremumSlip = 1.0f;
                LongitudalFunction.ExtremumValue = 0.02f;
                LongitudalFunction.AsymptoteSlip = 2.0f;
                LongitudalFunction.AsymptoteValue = 0.01f;
                LongitudalFunction.StiffnessFactor = 10000000.0f;

                // Almost no friction laterally
                TireForceFunctionDescription LateralFunction = new TireForceFunctionDescription();
                LateralFunction.ExtremumSlip = 1.0f;
                LateralFunction.ExtremumValue = 0.02f;
                LateralFunction.AsymptoteSlip = 2.0f;
                LateralFunction.AsymptoteValue = 0.01f;
                LateralFunction.StiffnessFactor = 0.0000001f;


                wheel0Prop.TireLongitudalForceFunction = LongitudalFunction;
                wheel1Prop.TireLongitudalForceFunction = LongitudalFunction;
                wheel2Prop.TireLongitudalForceFunction = LongitudalFunction;

                wheel0Prop.TireLateralForceFunction = LateralFunction;
                wheel1Prop.TireLateralForceFunction = LateralFunction;
                wheel2Prop.TireLateralForceFunction = LateralFunction;

                _wheel0 = new RobotinoWheel0Entity(wheel0Prop);
                _wheel0.State.Name = base.State.Name + " Wheel0";
                _wheel0.Parent = this;
                _wheel0.State.Assets.Mesh = "RobotinoWheel.bos";
                //_wheel0.MeshScale = new Vector3(0.01f, 0.01f, 0.01f);
                _wheel0.MeshRotation = new Vector3(0, -30.0f, 0);
                _wheel0.Initialize(device, physicsEngine);

                _wheel1 = new RobotinoWheel1Entity(wheel1Prop);
                _wheel1.State.Name = base.State.Name + " Wheel1";
                _wheel1.Parent = this;
                _wheel1.State.Assets.Mesh = "RobotinoWheel.bos";
                //_wheel1.MeshScale = new Vector3(0.01f, 0.01f, 0.01f);
                _wheel1.MeshRotation = new Vector3(0, 90, 0);
                _wheel1.Initialize(device, physicsEngine);

                _wheel2 = new RobotinoWheel2Entity(wheel2Prop);
                _wheel2.State.Name = base.State.Name + " Wheel2";
                _wheel2.Parent = this;
                _wheel2.State.Assets.Mesh = "RobotinoWheel.bos";
                //_wheel2.MeshScale = new Vector3(0.01f, 0.01f, 0.01f);
                _wheel2.MeshRotation = new Vector3(0, 30.0f, 0);
                _wheel2.Initialize(device, physicsEngine);

                State.Assets.Mesh = "Robotino.bos";
                //MeshScale = new Vector3( 0.01f, 0.01f, 0.01f );
                //MeshRotation = new Vector3(0, 180.0f, 0);

                base.Initialize(device, physicsEngine);
            }
            catch (Exception ex)
            {
                if (PhysicsEntity != null)
                    PhysicsEngine.DeleteEntity(PhysicsEntity);
                HasBeenInitialized = false;
                InitError = ex.ToString();
            }
        }

        public override void Render(RenderMode renderMode, MatrixTransforms transforms, CameraEntity currentCamera)
        {
            base.Render(renderMode, transforms, currentCamera);
            _wheel0.Render(renderMode, transforms, currentCamera);
            _wheel1.Render(renderMode, transforms, currentCamera);
            _wheel2.Render(renderMode, transforms, currentCamera);
        }
    }
}
