using System;
using System.Net;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Additional namespaces
using dssphttp = Microsoft.Dss.Core.DsspHttp;
using simengine = Microsoft.Robotics.Simulation.Engine;
using physics = Microsoft.Robotics.Simulation.Physics;
using contactsensor = Microsoft.Robotics.Services.ContactSensor.Proxy;
using robotino = RobotinoMRDS.Simulation.Robotino;
#endregion

namespace RobotinoMRDS.Simulation.Robotino.SimulatedRobotinoBumperArray
{
    [Contract(Contract.Identifier)]
    [DisplayName("SimulatedRobotinoBumperArray")]
    [Description("This service provides access to Robotino's Bumper Array")]
    [AlternateContract(contactsensor.Contract.Identifier)]
    class SimulatedRobotinoBumperArrayService : DsspServiceBase
    {
        #region Simulation variables
        robotino.RobotinoBumperArrayEntity _entity;
        simengine.SimulationEnginePort _notificationTarget;
        #endregion

        /// <summary>
        /// Service state
        /// </summary>
        contactsensor.ContactSensorArrayState _state = new contactsensor.ContactSensorArrayState();

        /// <summary>
        /// Entity Contact Notifications port
        /// </summary>
        Port<physics.EntityContactNotification> _contactNotificationPort = new Port<physics.EntityContactNotification>();

        /// <summary>
        /// Shapes to Sensor table
        /// </summary>
        Dictionary<physics.Shape, contactsensor.ContactSensor> _bumperShapeToSensorTable;

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/SimulatedRobotinoBumperArray", AllowMultipleInstances = true)]
        contactsensor.ContactSensorArrayOperations _mainPort = new contactsensor.ContactSensorArrayOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Service constructor
        /// </summary>
        public SimulatedRobotinoBumperArrayService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
                CreateDefaultState();
            _notificationTarget = new simengine.SimulationEnginePort();

            //PartnerType.Service is the entity instance name.
            simengine.SimulationEngine.GlobalInstancePort.Subscribe(ServiceInfo.PartnerList, _notificationTarget);

            // don't start listening to DSSP operations, other than drop,
            // until notification of entity
            Activate(new Interleave(
                new TeardownReceiverGroup
                (
                Arbiter.Receive<simengine.InsertSimulationEntity>(
                false,
                _notificationTarget,
                InsertEntityNotificationHandlerFirstTime),
                Arbiter.Receive<DsspDefaultDrop>(
                false,
                _mainPort,
                DefaultDropHandler)
                ),
                new ExclusiveReceiverGroup(),
                new ConcurrentReceiverGroup()
                ));
            base.Start();
        }

        private void CreateDefaultState()
        {
            _state.Sensors = new List<contactsensor.ContactSensor>();
            _bumperShapeToSensorTable = new Dictionary<physics.Shape, contactsensor.ContactSensor>();
        }

        void DeleteEntityNotificationHandler(simengine.DeleteSimulationEntity del)
        {
            _entity = null;
        }

        void InsertEntityNotificationHandlerFirstTime(simengine.InsertSimulationEntity ins)
        {
            InsertEntityNotificationHandler(ins);
            base.Start();
            MainPortInterleave.CombineWith(
                new Interleave(
                    new TeardownReceiverGroup(),
                    new ExclusiveReceiverGroup(
                        Arbiter.Receive<simengine.InsertSimulationEntity>(true, _notificationTarget, InsertEntityNotificationHandler),
                        Arbiter.Receive<simengine.DeleteSimulationEntity>(true, _notificationTarget, DeleteEntityNotificationHandler)
                    ),
                    new ConcurrentReceiverGroup()
                )
            );
        }

        void InsertEntityNotificationHandler(simengine.InsertSimulationEntity ins)
        {
            _entity = (robotino.RobotinoBumperArrayEntity)ins.Body;
            _entity.ServiceContract = Contract.Identifier;

            // reinitialize the state
            CreateDefaultState();

            contactsensor.ContactSensor cs = null;

            // The simulation bumper service uses a simple heuristic to assign contact sensors, to physics shapes:
            // Half the sensors go infront, the other half in the rear. Assume front sensors are first in the list
            // In the for loop below we create a lookup table that matches simulation shapes with sensors. When
            // the physics engine notifies us with a contact, it provides the shape that came into contact. We need to
            // translate that to sensor. In the future we might add an object Context field to shapes to make this easier
            for (int i = 0; i < _entity.Shapes.Length; i++)
            {
                cs = new contactsensor.ContactSensor();
                cs.Name = _entity.Shapes[i].State.Name;
                cs.HardwareIdentifier = i;
                _state.Sensors.Add(cs);
                _bumperShapeToSensorTable.Add((physics.BoxShape)_entity.Shapes[i], cs);
            }

            // subscribe to bumper simulation entity for contact notifications
            _entity.Subscribe(_contactNotificationPort);
            // Activate a handler on the notification port, it will run when contacts occur in simulation
            Activate(Arbiter.Receive(false, _contactNotificationPort, PhysicsContactNotificationHandler));
        }

        void PhysicsContactNotificationHandler(physics.EntityContactNotification contact)
        {
            foreach (physics.ShapeContact sc in contact.ShapeContacts)
            {
                // look up shape involved in collision and check its one of the bumper shapes
                contactsensor.ContactSensor s;
                if (!_bumperShapeToSensorTable.TryGetValue(sc.LocalShape, out s))
                    continue;
                if (contact.Stage == physics.ContactNotificationStage.Started)
                    s.Pressed = true;
                else if (contact.Stage == physics.ContactNotificationStage.Finished)
                    s.Pressed = false;
                s.TimeStamp = DateTime.Now;
                // notification for individual sensor
                _submgrPort.Post(new submgr.Submit(s, DsspActions.UpdateRequest));
            }

            // send notification
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // reactivate notification handler
            Activate(Arbiter.Receive(false, _contactNotificationPort, PhysicsContactNotificationHandler));
        }

        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(contactsensor.Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }


        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(dssphttp.HttpGet get)
        {
            get.ResponsePort.Post(new dssphttp.HttpResponseType(HttpStatusCode.OK, _state));
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(contactsensor.Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> SubscribeHandler(contactsensor.Subscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber, DsspActions.ReplaceRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
            );
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(contactsensor.ReliableSubscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber, DsspActions.ReplaceRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
            );
        }
    }
}


