using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Additional namespaces
using dssphttp = Microsoft.Dss.Core.DsspHttp;
using simengine = Microsoft.Robotics.Simulation.Engine;
using physics = Microsoft.Robotics.Simulation.Physics;
using drive = Microsoft.Robotics.Services.Drive.Proxy;
using robotino = RobotinoMRDS.Simulation.Robotino;
using xna = Microsoft.Xna.Framework;
using Microsoft.Robotics.PhysicalModel;
#endregion

namespace RobotinoMRDS.Simulation.RobotinoSimulatedRobotinoDrive
{
    [Contract(Contract.Identifier)]
    [DisplayName("SimulatedRobotinoDrive")]
    [Description("This service provides access to Robotino's drive system")]
    [AlternateContract(drive.Contract.Identifier)]
    class SimulatedRobotinoDriveService : DsspServiceBase
    {

        #region Simulation variables
        robotino.RobotinoEntity _entity;
        simengine.SimulationEnginePort _notificationTarget;
        #endregion

        /// <summary>
        /// Service state
        /// </summary>
        [InitialStatePartner(Optional = true, ServiceUri = "SimulatedRobotinoDriveService.Config.xml")]
        SimulatedRobotinoDriveState _state = new SimulatedRobotinoDriveState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/SimulatedRobotinoDrive", AllowMultipleInstances = true)]
        SimulatedRobotinoDriveOperations _mainPort = new SimulatedRobotinoDriveOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Alternate service port
        /// </summary>
        [AlternateServicePort(
            AllowMultipleInstances = true,
            AlternateContract = drive.Contract.Identifier)]
        private drive.DriveOperations _diffDrivePort = new drive.DriveOperations();

        /// <summary>
        /// Service constructor
        /// </summary>
        public SimulatedRobotinoDriveService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        #region Initialization and State
        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
                CreateDefaultState();
            _notificationTarget = new simengine.SimulationEnginePort();

            //PartnerType.Service is the entity instance name.
            simengine.SimulationEngine.GlobalInstancePort.Subscribe(ServiceInfo.PartnerList, _notificationTarget);

            // don't start listening to DSSP operations, other than drop,
            // until notification of entity
            Activate(new Interleave(
                new TeardownReceiverGroup
                (
                Arbiter.Receive<simengine.InsertSimulationEntity>(
                false,
                _notificationTarget,
                InsertEntityNotificationHandlerFirstTime),
                Arbiter.Receive<DsspDefaultDrop>(
                false,
                _diffDrivePort,
                DefaultDropHandler),
                Arbiter.Receive<DsspDefaultDrop>(
                false,
                _mainPort,
                DefaultDropHandler)
                ),
                new ExclusiveReceiverGroup(),
                new ConcurrentReceiverGroup()
                ));
        }
        void CreateDefaultState()
        {
            _state = new SimulatedRobotinoDriveState();
            _state.LeftWheel = new Microsoft.Robotics.Services.Motor.Proxy.WheeledMotorState();
            _state.RightWheel = new Microsoft.Robotics.Services.Motor.Proxy.WheeledMotorState();
            _state.Wheel1 = new Microsoft.Robotics.Services.Motor.Proxy.WheeledMotorState();

            _state.LeftWheel.MotorState = new Microsoft.Robotics.Services.Motor.Proxy.MotorState();
            _state.RightWheel.MotorState = new Microsoft.Robotics.Services.Motor.Proxy.MotorState();
            _state.Wheel1.MotorState = new Microsoft.Robotics.Services.Motor.Proxy.MotorState();

            _state.LeftWheel.EncoderState = new Microsoft.Robotics.Services.Encoder.Proxy.EncoderState();
            _state.RightWheel.EncoderState = new Microsoft.Robotics.Services.Encoder.Proxy.EncoderState();
            _state.Wheel1.EncoderState = new Microsoft.Robotics.Services.Encoder.Proxy.EncoderState();
        }

        void UpdateStateFromSimulation()
        {
            if (_entity != null)
            {
                _state.TimeStamp = DateTime.Now;
                _state.LeftWheel.MotorState.CurrentPower = _entity.Wheel0.Wheel.MotorTorque;
                _state.LeftWheel.EncoderState.CurrentReading = _entity.Wheel0Encoder;
                _state.RightWheel.MotorState.CurrentPower = _entity.Wheel2.Wheel.MotorTorque;
                _state.RightWheel.EncoderState.CurrentReading = _entity.Wheel2Encoder;
                _state.Wheel1.MotorState.CurrentPower = _entity.Wheel1.Wheel.MotorTorque;
                _state.Wheel1.EncoderState.CurrentReading = _entity.Wheel1Encoder;
                _state.Position = _entity.State.Pose.Position;
            }
        }
        #endregion

        #region Entity handlers

        void InsertEntityNotificationHandlerFirstTime(simengine.InsertSimulationEntity ins)
        {
            //insert overselves into the directory
            base.Start();

            InsertEntityNotificationHandler(ins);
        }

        void InsertEntityNotificationHandler(simengine.InsertSimulationEntity ins)
        {
            _entity = (robotino.RobotinoEntity)ins.Body;
            _entity.ServiceContract = Contract.Identifier;

            // create default state based on the physics entity
            xna.Vector3 separation = _entity.Wheel0.Position - _entity.Wheel2.Position;
            _state.DistanceBetweenWheels = separation.Length();

            _state.LeftWheel.MotorState.PowerScalingFactor = _entity.MotorTorqueScaling;
            _state.RightWheel.MotorState.PowerScalingFactor = _entity.MotorTorqueScaling;

            //enable other handlers now that we are connected
            Activate(new Interleave(
                new TeardownReceiverGroup
                (
                Arbiter.Receive<DsspDefaultDrop>(false, _mainPort, DefaultDropHandler),
                Arbiter.Receive<DsspDefaultDrop>(false, _diffDrivePort, DefaultDropHandler)
                ),
                new ExclusiveReceiverGroup
                (
                Arbiter.Receive<SetPose>(true, _mainPort, SetPoseHandler),
                Arbiter.ReceiveWithIterator<drive.DriveDistance>(
                true, _diffDrivePort, DriveDistanceHandler),
                Arbiter.ReceiveWithIterator<drive.RotateDegrees>(
                true, _diffDrivePort, RotateHandler),
                Arbiter.ReceiveWithIterator<drive.SetDrivePower>(
                true, _diffDrivePort, SetPowerHandler),
                Arbiter.ReceiveWithIterator<drive.SetDriveSpeed>(
                true, _diffDrivePort, SetSpeedHandler),
                Arbiter.ReceiveWithIterator<drive.AllStop>(
                true, _diffDrivePort, AllStopHandler),
                Arbiter.ReceiveWithIterator<OmniDrive>(
                true, _mainPort, OmniDriveHandler),
                Arbiter.ReceiveWithIterator<Rotate>(
                true, _mainPort, RotateHandler),
                Arbiter.Receive<simengine.InsertSimulationEntity>(
                true, _notificationTarget, InsertEntityNotificationHandler),
                Arbiter.Receive<simengine.DeleteSimulationEntity>(
                true, _notificationTarget, DeleteEntityNotificationHandler)
                ),
                new ConcurrentReceiverGroup
                (
                Arbiter.ReceiveWithIterator<dssphttp.HttpGet>(
                true, _mainPort, MainPortHttpGetHandler),
                Arbiter.ReceiveWithIterator<Get>(true, _mainPort, MainPortGetHandler),
                Arbiter.ReceiveWithIterator<dssphttp.HttpGet>(
                true, _diffDrivePort, HttpGetHandler),
                Arbiter.ReceiveWithIterator<drive.Get>(true, _diffDrivePort, GetHandler),
                Arbiter.ReceiveWithIterator<drive.Subscribe>(true, _diffDrivePort, SubscribeHandler),
                Arbiter.ReceiveWithIterator<drive.ReliableSubscribe>(
                true, _diffDrivePort, ReliableSubscribeHandler),
                Arbiter.ReceiveWithIterator<drive.EnableDrive>(true, _diffDrivePort, EnableHandler)
                )
                ));
        }
        void DeleteEntityNotificationHandler(simengine.DeleteSimulationEntity del)
        {
            _entity = null;

            // disable other handlers now that we are not longer connected to the entity
            Activate(new Interleave(
                new TeardownReceiverGroup(
                    Arbiter.Receive<simengine.InsertSimulationEntity>(
                    true, _notificationTarget, InsertEntityNotificationHandlerFirstTime),
                    Arbiter.Receive<DsspDefaultDrop>(
                    true, _mainPort, DefaultDropHandler),
                    Arbiter.Receive<DsspDefaultDrop>(
                    true, _diffDrivePort, DefaultDropHandler)
                    ),
                    new ExclusiveReceiverGroup(),
                    new ConcurrentReceiverGroup()
                    ));
        }
        #endregion

        #region Subscribe handlers
        public IEnumerator<ITask> SubscribeHandler(
            drive.Subscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(
                _submgrPort,
                subscribe.Body,
                subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber,
                        DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
                ));
            yield break;
        }

        public IEnumerator<ITask> ReliableSubscribeHandler(
            drive.ReliableSubscribe subscribe)
        {
            Activate(Arbiter.Choice(
                SubscribeHelper(
                _submgrPort,
                subscribe.Body,
                subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _submgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber,
                        DsspActions.UpdateRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
                ));
            yield break;
        }
        #endregion

        #region MainPort handlers

        public IEnumerator<ITask> MainPortHttpGetHandler(dssphttp.HttpGet get)
        {
            UpdateStateFromSimulation();
            get.ResponsePort.Post(new dssphttp.HttpResponseType(_state));
            yield break;
        }

        public IEnumerator<ITask> MainPortGetHandler(Get get)
        {
            UpdateStateFromSimulation();
            get.ResponsePort.Post(_state);
            yield break;
        }

        public void SetPoseHandler(SetPose setPose)
        {
            if (_entity == null)
                throw new InvalidOperationException(
                    "Simulation entity not registered with service");
            Task<robotino.RobotinoEntity, Pose> task = new
                Task<robotino.RobotinoEntity, Pose>(_entity, setPose.Body.EntityPose, SetPoseDeferred);

            _entity.DeferredTaskQueue.Post(task);
        }

        void SetPoseDeferred(robotino.RobotinoEntity entity, Pose pose)
        {
            if (_entity == null)
                throw new InvalidOperationException(
                    "Simulation entity not registered with service");

            entity.PhysicsEntity.SetPose(pose);
        }

        public IEnumerator<ITask> OmniDriveHandler(OmniDrive omdri)
        {
            if (_entity == null)
                throw new InvalidOperationException(
                    "Simulation entity not registered with service");

            SuccessFailurePort entityResponse = new SuccessFailurePort();
            _entity.OmniDrive(
                (float)omdri.Body.VelocityX,
                (float)omdri.Body.VelocityZ,
                (float)omdri.Body.Omega,
                entityResponse);
            yield return Arbiter.Choice(entityResponse,
                delegate(SuccessResult s)
                {
                    omdri.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                },
                delegate(Exception e)
                {
                    omdri.ResponsePort.Post(new W3C.Soap.Fault());
                });
            yield break;
        }

        public IEnumerator<ITask> RotateHandler(Rotate rot)
        {
            if (_entity == null)
                throw new InvalidOperationException(
                    "Simulation entity not registered with service");

            SuccessFailurePort entityResponse = new SuccessFailurePort();
            _entity.Rotate(
                rot.Body.Direction,
                (float)rot.Body.Power,
                entityResponse);

            yield return Arbiter.Choice(entityResponse,
               delegate(SuccessResult s)
               {
                   rot.ResponsePort.Post(DefaultUpdateResponseType.Instance);
               },
               delegate(Exception e)
               {
                   rot.ResponsePort.Post(new W3C.Soap.Fault());
               });
            yield break;

        }
        #endregion

        #region _diffDrivePort handlers

        public IEnumerator<ITask> HttpGetHandler(dssphttp.HttpGet get)
        {
            UpdateStateFromSimulation();
            drive.DriveDifferentialTwoWheelState _twoWheelState =
                (drive.DriveDifferentialTwoWheelState)
                ((drive.DriveDifferentialTwoWheelState)_state).Clone();
            get.ResponsePort.Post(new dssphttp.HttpResponseType(
                _twoWheelState));
            yield break;
        }

        public IEnumerator<ITask> GetHandler(drive.Get get)
        {
            UpdateStateFromSimulation();

            drive.DriveDifferentialTwoWheelState _twoWheelState =
                (drive.DriveDifferentialTwoWheelState)
                ((drive.DriveDifferentialTwoWheelState)_state).Clone();
            get.ResponsePort.Post(_twoWheelState);
            yield break;
        }

        public IEnumerator<ITask> DriveDistanceHandler(drive.DriveDistance driveDistance)
        {
            SuccessFailurePort entityResponse = new SuccessFailurePort();
            _entity.DriveDistance(
                (float)driveDistance.Body.Distance,
                (float)driveDistance.Body.Power,
                entityResponse);

            yield return Arbiter.Choice(entityResponse,
                delegate(SuccessResult s)
                {
                    driveDistance.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                },
                delegate(Exception e)
                {
                    driveDistance.ResponsePort.Post(new W3C.Soap.Fault());
                });
            yield break;
        }

        public IEnumerator<ITask> RotateHandler(drive.RotateDegrees rotateDegrees)
        {
            SuccessFailurePort entityResponse = new SuccessFailurePort();
            _entity.RotateDegrees(
                (float)rotateDegrees.Body.Degrees,
                (float)rotateDegrees.Body.Power,
                entityResponse);

            yield return Arbiter.Choice(entityResponse,
                delegate(SuccessResult s)
                {
                    rotateDegrees.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                },
                delegate(Exception e)
                {
                    rotateDegrees.ResponsePort.Post(new W3C.Soap.Fault());
                });
            yield break;
        }

        public IEnumerator<ITask> SetPowerHandler(drive.SetDrivePower setPower)
        {
            if (_entity == null)
                throw new InvalidOperationException("Simulation entity not registered with service");

            // Call simulation entity method for setting wheel torque
            _entity.SetMotorTorque(
                (float)(setPower.Body.LeftWheelPower),
                0.0f,
                (float)(setPower.Body.RightWheelPower));

            UpdateStateFromSimulation();
            setPower.ResponsePort.Post(DefaultUpdateResponseType.Instance);

            // send update notification for entire state
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.UpdateRequest));
            yield break;
        }

        public IEnumerator<ITask> SetSpeedHandler(drive.SetDriveSpeed setSpeed)
        {
            if (_entity == null)
                throw new InvalidOperationException("Simulation entity not registered with service");

            _entity.SetVelocity(
                (float)setSpeed.Body.LeftWheelSpeed,
                0.0f,
                (float)setSpeed.Body.RightWheelSpeed);

            UpdateStateFromSimulation();
            setSpeed.ResponsePort.Post(DefaultUpdateResponseType.Instance);

            // send update notification for entire state
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.UpdateRequest));
            yield break;
        }

        public IEnumerator<ITask> EnableHandler(drive.EnableDrive enable)
        {
            if (_entity == null)
                throw new InvalidOperationException("Simulation entity not registered with service");

            _state.IsEnabled = enable.Body.Enable;
            _entity.IsEnabled = _state.IsEnabled;

            UpdateStateFromSimulation();

            enable.ResponsePort.Post(DefaultUpdateResponseType.Instance);

            // send update notification for entire state
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.UpdateRequest));
            yield break;
        }

        public IEnumerator<ITask> AllStopHandler(drive.AllStop allstop)
        {
            if (_entity == null)
                throw new InvalidOperationException("Simulation entity not registered with service");

            _entity.SetMotorTorque(0, 0, 0);
            _entity.SetVelocity(0);

            UpdateStateFromSimulation();
            Console.WriteLine("All stop executed!");

            allstop.ResponsePort.Post(DefaultUpdateResponseType.Instance);

            // send update notification for entire state
            _submgrPort.Post(new submgr.Submit(_state, DsspActions.UpdateRequest));
            yield break;
        }

        #endregion
    }
}