using System;
using System.Net;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Additional namespaces
using dssphttp = Microsoft.Dss.Core.DsspHttp;
using simengine = Microsoft.Robotics.Simulation.Engine;
using physics = Microsoft.Robotics.Simulation.Physics;
using analogsensorarray = Microsoft.Robotics.Services.AnalogSensorArray.Proxy;
using analogsensor = Microsoft.Robotics.Services.AnalogSensor.Proxy;
using robotino = RobotinoMRDS.Simulation.Robotino;
#endregion

namespace RobotinoMRDS.Simulation.Robotino.SimulatedRobotinoIRArray
{
    [Contract(Contract.Identifier)]
    [DisplayName("SimulatedRobotinoIRArray")]
    [Description("This service provides access to Robotino's IR Array")]
    [AlternateContract(analogsensorarray.Contract.Identifier)]
    class SimulatedRobotinoIRArrayService : DsspServiceBase
    {
        #region Simulation variables
        robotino.RobotinoIRArrayEntity _entity;
        simengine.SimulationEnginePort _notificationTarget;
        #endregion

        float[] _previousDistances = new float[] {0f, 0f, 0f,
                                     0f, 0f, 0f,
                                     0f, 0f, 0f};


        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        analogsensorarray.AnalogSensorArrayState _state = new analogsensorarray.AnalogSensorArrayState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/SimulatedRobotinoIRArray", AllowMultipleInstances = true)]
        analogsensorarray.AnalogSensorOperations _mainPort = new analogsensorarray.AnalogSensorOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Service constructor
        /// </summary>
        public SimulatedRobotinoIRArrayService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            _notificationTarget = new simengine.SimulationEnginePort();

            // PartnerType.Service is the entity instance name. 
            simengine.SimulationEngine.GlobalInstancePort.Subscribe(ServiceInfo.PartnerList, _notificationTarget);

            // dont start listening to DSSP operations, other than drop, until notification of entity
            Activate(new Interleave(
                new TeardownReceiverGroup
                (
                    Arbiter.Receive<simengine.InsertSimulationEntity>(false, _notificationTarget, InsertEntityNotificationHandlerFirstTime),
                    Arbiter.Receive<DsspDefaultDrop>(false, _mainPort, DefaultDropHandler)
                ),
                new ExclusiveReceiverGroup(),
                new ConcurrentReceiverGroup()
            ));

            // start notification method
            SpawnIterator<DateTime>(DateTime.Now, CheckForStateChange);
        }

        /// <summary>
        /// Check the hardware value 5 times per second and send a notification if it has changed.
        /// </summary>
        /// <param name="timeout"></param>
        /// <returns></returns>
        private IEnumerator<ITask> CheckForStateChange(DateTime timeout)
        {
            while (true)
            {
                if (_entity != null)
                {
                    robotino.RobotinoIREntity[] _sensors = _entity.IRSensors;
                    for (int i = 0; i < _sensors.Length; i++)
                    {
                        if (_sensors[i].Distance != _previousDistances[i])
                        {
                            // send notification of state change
                            UpdateState();
                            base.SendNotification<analogsensorarray.Replace>(_submgrPort, _state);
                        }
                        _previousDistances[i] = _sensors[i].Distance;
                    }
                }
                yield return Arbiter.Receive(false, TimeoutPort(200), delegate { });
            }
        }

        private void CreateDefaultState()
        {
            _state.Sensors = new List<analogsensor.AnalogSensorState>();
            analogsensor.AnalogSensorState ir = null;

            for (int i = 0; i < _entity.IRSensors.Length; i++)
            {
                ir = new analogsensor.AnalogSensorState();
                ir.HardwareIdentifier = i;
                ir.RawMeasurement = 0.0f;
                ir.RawMeasurementRange = 0.0f;
                ir.NormalizedMeasurement = 0.0f;
                ir.TimeStamp = DateTime.Now;
                _state.Sensors.Add(ir);
            }
        }

        void DeleteEntityNotificationHandler(simengine.DeleteSimulationEntity del)
        {
            _entity = null;
        }

        /// <summary>
        /// Called the first time the simulation engine tells us about our entity.  Activate
        /// the other handlers and insert ourselves into the service directory.
        /// </summary>
        /// <param name="ins"></param>
        void InsertEntityNotificationHandlerFirstTime(simengine.InsertSimulationEntity ins)
        {
            InsertEntityNotificationHandler(ins);
            base.Start();
            MainPortInterleave.CombineWith(
                new Interleave(
                    new TeardownReceiverGroup(),
                    new ExclusiveReceiverGroup(
                        Arbiter.Receive<simengine.InsertSimulationEntity>(true, _notificationTarget, InsertEntityNotificationHandler),
                        Arbiter.Receive<simengine.DeleteSimulationEntity>(true, _notificationTarget, DeleteEntityNotificationHandler)
                    ),
                    new ConcurrentReceiverGroup()
                )
            );
        }

        void InsertEntityNotificationHandler(simengine.InsertSimulationEntity ins)
        {
            _entity = (robotino.RobotinoIRArrayEntity)ins.Body;
            _entity.ServiceContract = Contract.Identifier;

            CreateDefaultState();
        }

        void UpdateState()
        {
            robotino.RobotinoIREntity[] _sensors = _entity.IRSensors;
            // update our state from the entity
            for (int i = 0; i < _state.Sensors.Count; i++)
            {
                _state.Sensors[i].RawMeasurement = _sensors[i].Distance;
                _state.Sensors[i].RawMeasurementRange = _sensors[i].MaximumRange;
                _state.Sensors[i].NormalizedMeasurement = _state.Sensors[i].RawMeasurement / _state.Sensors[i].RawMeasurementRange;
                _state.Sensors[i].TimeStamp = DateTime.Now;
            }
        }

/// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(analogsensorarray.Get get)
        {
            UpdateState();
            get.ResponsePort.Post(_state);
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(dssphttp.HttpGet get)
        {
            UpdateState();
            get.ResponsePort.Post(new dssphttp.HttpResponseType(HttpStatusCode.OK, _state));
            yield break;
        }

        /// <summary>
        /// Replace Handler
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(analogsensorarray.Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }
        /// <summary>
        /// ReliableSubscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(analogsensorarray.ReliableSubscribe subscribe)
        {
            yield break;
        }
        /// <summary>
        /// Subscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(analogsensorarray.Subscribe subscribe)
        {
            SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }
    }
}


