using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

#region Additional Namespaces
using xna = Microsoft.Xna.Framework;
using engineproxy = Microsoft.Robotics.Simulation.Engine.Proxy;
using Microsoft.Robotics.Simulation;
using Microsoft.Robotics.Simulation.Physics;
using Microsoft.Robotics.Simulation.Engine;
using Microsoft.Robotics.PhysicalModel;
using robotino = RobotinoMRDS.Simulation.Robotino;
#endregion
namespace RobotinoMRDS.Simulation.Scenarios.WallFollow.PopulateEnvironment
{
    [Contract(Contract.Identifier)]
    [DisplayName("PopulateEnvironment")]
    [Description("PopulateEnvironment service (no description provided)")]
    class PopulateEnvironmentService : DsspServiceBase
    {
        // Partner with the Simulation Engine
        [Partner(
            "Engine",
            Contract = engineproxy.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        engineproxy.SimulationEnginePort _engineStub = new engineproxy.SimulationEnginePort();

        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        PopulateEnvironmentState _state = new PopulateEnvironmentState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/PopulateEnvironment", AllowMultipleInstances = true)]
        PopulateEnvironmentOperations _mainPort = new PopulateEnvironmentOperations();

        /// <summary>
        /// Service constructor
        /// </summary>
        public PopulateEnvironmentService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            base.Start();
            Populate();
        }
        
        private void Populate()
        {
            // Set up initial view
            CameraView view = new CameraView();
            view.EyePosition = new Vector3(-0.91f, 0.67f, -1f);
            view.LookAtPoint = new Vector3(1.02f, 0.09f, 0.19f);
            SimulationEngine.GlobalInstancePort.Update(view);

            // Add another camera to view the scene from above
            CameraEntity fromAbove = new CameraEntity();
            fromAbove.State.Name = "FromAbove";
            fromAbove.Location = new xna.Vector3(-2.12f, 9.27f, -0.45f);
            fromAbove.LookAt = new xna.Vector3(-2.1f, -1.78f, -1.45f);
            fromAbove.IsRealTimeCamera = false;
            SimulationEngine.GlobalInstancePort.Insert(fromAbove);

            // Add a SkyDome.
            SkyDomeEntity sky = new SkyDomeEntity("skydome.dds", "sky_diff.dds");
            SimulationEngine.GlobalInstancePort.Insert(sky);

            // Add a directional light to simulate the sun.
            LightSourceEntity sun = new LightSourceEntity();
            sun.State.Name = "Sun";
            sun.Type = LightSourceEntityType.Directional;
            sun.Color = new Vector4(1, 1, 1, 1);
            sun.Direction = new Vector3(-0.47f, -0.8f, -0.36f);
            SimulationEngine.GlobalInstancePort.Insert(sun);

            HeightFieldShapeProperties hf = new HeightFieldShapeProperties("height field",
                64, // number of rows 
                10, // distance in meters, between rows
                64, // number of columns
                10, // distance in meters, between columns
                1, // scale factor to multiple height values 
                -1000); // vertical extent of the height field. Should be set to large negative values

            // create array with height samples
            hf.HeightSamples = new HeightFieldSample[hf.RowCount * hf.ColumnCount];
            for (int i = 0; i < hf.RowCount * hf.ColumnCount; i++)
            {
                hf.HeightSamples[i] = new HeightFieldSample();
                hf.HeightSamples[i].Height = (short)(Math.Sin(i * 0.01));
            }

            // create a material for the entire field. We could also specify material per sample.
            hf.Material = new MaterialProperties("ground", 0.8f, 0.5f, 0.8f);

            // insert ground entity in simulation and specify a texture
            SimulationEngine.GlobalInstancePort.Insert(new HeightFieldEntity(hf, "univ2.dds"));

            Wall[] Walls = new Wall[]
            {
                //new Wall("Wall0", new Vector3(5.5f, 0, -4), new Vector3(4f, 0.8f, 0.1f), new Quaternion(0,0,0,1)),
                //new Wall("Wall1", new Vector3(3.55f, 0, -3.05f), new Vector3(2f, 0.8f, 0.1f), Quaternion.FromAxisAngle(0, 1, 0, (float)(Math.PI / 2))),
                new Wall("Wall0", new Vector3(2f, 0, 0), new Vector3(4f, 0.8f, 0.1f), Quaternion.FromAxisAngle(0, 1, 0, (float)(Math.PI / 2))),
                new Wall("Wall1", new Vector3(0f, 0, -2.05f), new Vector3(4f, 0.8f, 0.1f), new Quaternion(0,0,0,1)),
                new Wall("Wall2", new Vector3(-2f, 0, 0), new Vector3(4f, 0.8f, 0.1f), Quaternion.FromAxisAngle(0, 1, 0, (float)(Math.PI / 2))),
                new Wall("Wall3", new Vector3(0f, 0, 2), new Vector3(4f, 0.8f, 0.1f), new Quaternion(0,0,0,1)),

            };

            foreach (Wall w in Walls)
            {
                SingleShapeEntity wall =
                    new SingleShapeEntity(
                        new BoxShape(
                            new BoxShapeProperties(
                                0, //no mass makes a static shape
                                new Pose(),
                                w.Dimensions)), //dimensions
                                w.Position);
                wall.State.Pose.Orientation = w.Orientation;
                wall.State.Name = w.Name;
                wall.State.Assets.DefaultTexture = "MayangConcrP.dds";
                SimulationEngine.GlobalInstancePort.Insert(wall);
            }
            // Finally insert Robotino
            SimulationEngine.GlobalInstancePort.Insert(
                new robotino.RobotinoEntity(
                    "Robotino", 
                    new Pose(
                        new Vector3(0, 0, 0),
                        Quaternion.FromAxisAngle(0, 1, 0, (float)Math.PI)
                        )));
        }
    }
    public struct Wall
    {
        public string Name;
        public Vector3 Position;
        public Vector3 Dimensions;
        public Quaternion Orientation;
        public Wall(string name, Vector3 position, Vector3 dimensions, Quaternion orientation)
        {
            Name = name;
            Position = position;
            Dimensions = dimensions;
            Orientation = orientation;
        }
    }
}


