using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

#region Additional Namespaces
using drive = Microsoft.Robotics.Services.Drive.Proxy;
using omniDrive = RobotinoMRDS.Simulation.RobotinoSimulatedRobotinoDrive.Proxy;
using irsensorarray = Microsoft.Robotics.Services.AnalogSensorArray.Proxy;
using env = RobotinoMRDS.Simulation.Scenarios.WallFollow.PopulateEnvironment.Proxy;
using xna = Microsoft.Xna.Framework;
using robotino = RobotinoMRDS.Simulation.Robotino;
using System.Text;
#endregion

namespace RobotinoMRDS.Simulation.Scenarios.WallFollow.SimWallFollow
{
    [Contract(Contract.Identifier)]
    [DisplayName("SimWallFollow")]
    [Description("A service that implements wall following using the readings from Robotino's IR Sensors")]
    class SimWallFollowService : DsspServiceBase
    {
        // used to communicate with the SimulatedRobotinoDrive service
        [Partner(
            "OmniDrive",
            Contract = omniDrive.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExisting,
            Optional = false)]
        omniDrive.SimulatedRobotinoDriveOperations _omniDrivePort = new omniDrive.SimulatedRobotinoDriveOperations();

        // used to communicate with the DifferentialDrive alternate
        [Partner(
            "Drive",
            Contract = drive.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExisting,
            Optional = false)]
        drive.DriveOperations _drivePort = new drive.DriveOperations();

        // used to communicate with the IR sensor service
        [Partner(
            "IRSensorArray",
            Contract = irsensorarray.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExisting,
            Optional = false)]
        irsensorarray.AnalogSensorOperations _irSensorArrayPort = new irsensorarray.AnalogSensorOperations();
        irsensorarray.AnalogSensorOperations _irNotifyPort = new irsensorarray.AnalogSensorOperations();

        // used to communicate with the referee service
        [Partner(
            "Environment",
            Contract = env.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExisting,
            Optional = false)]
        env.PopulateEnvironmentOperations _refereePort = new env.PopulateEnvironmentOperations();

        const float SLOW_VELOCITY = 0.5f;
        const float MEDIUM_VELOCITY = 1.0f;
        const float VELOCITY = 2.0f;
        const float FAST_VELOCITY = 2.5f;
        const float ANGULARVELOCITY = 0.02f;

        /// <summary>
        /// Service state
        /// </summary>
        [ServiceState]
        SimWallFollowState _state = new SimWallFollowState();

        /// <summary>
        /// Main service port
        /// </summary>
        [ServicePort("/SimWallFollow", AllowMultipleInstances = true)]
        SimWallFollowOperations _mainPort = new SimWallFollowOperations();

        /// <summary>
        /// Service constructor
        /// </summary>
        public SimWallFollowService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            base.Start();
            SpawnIterator<DateTime>(DateTime.Now, FollowWalls);
        }

        IEnumerator<ITask> FollowWalls(DateTime timeout)
        {
            _irSensorArrayPort.Subscribe(_irNotifyPort);
            _drivePort.EnableDrive(true);

            float[] ev = new float[] { 1.0f, 0.0f };

            // escape vector for distance sensor
            float[][] escapeVector = new float[][]
            {
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f},
                new float[]{0.0f, 0.0f}
            };

            for (int i = 0; i < 9; i++)
            {
                rotate(ev, escapeVector[i], 40.0f * i);
            }
            const float ESCAPE_DISTANCE = 0.10f;
            const float WALL_LOST_DISTANCE = 0.35f;
            const float WALL_FOUND_DISTANCE = 0.30f;
            const float WALL_FOLLOW_DISTANCE = 0.15f;
            const float NEW_WALL_FOUND_DISTANCE = 0.12f;
            float[] escape = new float[] { 0.0f, 0.0f };
            int curWallSensor = -1;
            float[] dir = new float[] { 1.0f, 0.0f };
            float velocity = VELOCITY;
            float rotVelocity = ANGULARVELOCITY;

            irsensorarray.AnalogSensorArrayState infrared = new irsensorarray.AnalogSensorArrayState();

            while (true)
            {
                velocity = VELOCITY;
                int minIndex = 0;
                float minDistance = 0.40f;
                int numEscape = 0;
                escape[0] = escape[1] = 0.0f;

                // get the IR sensor reading
                irsensorarray.Get tmp = new irsensorarray.Get();
                _irSensorArrayPort.Post(tmp);

                yield return Arbiter.Choice(tmp.ResponsePort,
                    delegate(irsensorarray.AnalogSensorArrayState state) { infrared = state; },
                    delegate(Fault f) { Console.WriteLine(f); }
                    );

                StringBuilder values = new StringBuilder();
                for (int i = 0; i < infrared.Sensors.Count; ++i)
                {
                    float v = (float)infrared.Sensors[i].RawMeasurement;
                    values.Append(v.ToString() + " ");
                    if (v < minDistance)
                    {
                        minDistance = v;
                        minIndex = i;
                    }
                    if (v < ESCAPE_DISTANCE)
                    {
                        ++numEscape;
                        addScaledVector(escape, escapeVector[i], v);
                    }
                }
                Console.WriteLine(values.ToString());
                if (numEscape >= 2)
                {
                    // close to walls with more than one sensor, try to escape
                    normalizeVector(escape);
                    rotate(escape, dir, 180);
                    velocity = SLOW_VELOCITY;
                }
                else
                {
                    if (curWallSensor != -1
                    && infrared.Sensors[curWallSensor].RawMeasurement > WALL_LOST_DISTANCE)
                    {
                        curWallSensor = -1;
                        rotateInPlace(dir, -20);
                        rotVelocity = -0.20f;
                    }
                    if (curWallSensor == -1)
                    {
                        // wall not found yet
                        if (minDistance < WALL_FOUND_DISTANCE)
                        {
                            curWallSensor = minIndex;
                            velocity = SLOW_VELOCITY;
                        }
                    }
                    if (curWallSensor != -1)
                    {
                        float wallDist = (float)infrared.Sensors[curWallSensor].RawMeasurement;

                        // check for global new wall
                        if (minIndex != curWallSensor && minDistance < NEW_WALL_FOUND_DISTANCE)
                        {
                            // new wall found, drive along this wall
                            curWallSensor = minIndex;
                            wallDist = (float)infrared.Sensors[curWallSensor].RawMeasurement;
                            velocity = SLOW_VELOCITY;
                        }
                        else
                        {
                            if (infrared.Sensors[(curWallSensor + 1) % 9].RawMeasurement < wallDist)
                            {
                                // switch walls
                                curWallSensor = (curWallSensor + 1) % 9;
                                velocity = MEDIUM_VELOCITY;
                            }
                            // check for new wall in direction
                            for (int i = 0; i < 2; ++i)
                            {
                                int tmpId = (curWallSensor + 2 + i) % 9;
                                if (infrared.Sensors[tmpId].RawMeasurement < WALL_FOUND_DISTANCE)
                                {
                                    curWallSensor = tmpId;
                                    wallDist = (float)infrared.Sensors[tmpId].RawMeasurement;
                                    velocity = SLOW_VELOCITY;
                                    break;
                                }
                            }
                        }
                        // try to keep neighbor distance sensors in balance
                        float vr = (float)infrared.Sensors[(curWallSensor + 1) % 9].RawMeasurement;
                        float vl = (float)infrared.Sensors[(curWallSensor + 8) % 9].RawMeasurement;

                        rotVelocity = (vr - vl);
                        float followAngle = 95;
                        if (Math.Abs(rotVelocity) > 0.30)
                        {
                            velocity = SLOW_VELOCITY;
                            followAngle = rotVelocity >= 0.0 ? 140 : 80;
                        }
                        else if (Math.Abs(rotVelocity) > 0.40)
                        {
                            velocity = MEDIUM_VELOCITY;
                            followAngle = rotVelocity >= 0.0 ? 120 : 85;
                        }
                        rotVelocity *= 8 * ANGULARVELOCITY;
                        // follow the wall to the left
                        rotate(escapeVector[curWallSensor], dir, followAngle);
                        // keep distance to wall steady
                        float scale = wallDist - WALL_FOLLOW_DISTANCE;

                        //if (scale > 0)
                        //{
                            scale *= 10f;
                        //}
                        addScaledVector(dir, escapeVector[curWallSensor], scale);
                        normalizeVector(dir);
                    }
                }
                if (minDistance > 0.20f)
                {
                    velocity = FAST_VELOCITY;
                }
               yield return Arbiter.Choice(
                    _omniDrivePort.OmniDrive(velocity * (float)dir[0], velocity * (float)dir[1], rotVelocity),
                    delegate(DefaultUpdateResponseType response) { },
                    delegate(Fault f) { LogError(f); }
                );
                yield return Arbiter.Receive(false, TimeoutPort(100), delegate { });
            }
        }
        void rotate(float[] inArray, float[] outArray, float a)
        {
            float rad = xna.MathHelper.ToRadians(a);
            outArray[0] = (float)Math.Cos(rad) * inArray[0] - (float)Math.Sin(rad) * inArray[1];
            outArray[1] = (float)Math.Sin(rad) * inArray[0] + (float)Math.Cos(rad) * inArray[1];
        }
        void rotateInPlace(float[] v, float a)
        {
            float rad = xna.MathHelper.ToRadians(a);
            float tmp = v[0];
            v[0] = (float)Math.Cos(rad) * v[0] - (float)Math.Sin(rad) * v[1];
            v[1] = (float)Math.Sin(rad) * tmp + (float)Math.Cos(rad) * v[1];
        }
        void addScaledVector(float[] srcDest, float[] uv, float scale)
        {
            srcDest[0] += uv[0] * scale;
            srcDest[1] += uv[1] * scale;
        }
        void normalizeVector(float[] v)
        {
            float len = (float)Math.Sqrt(v[0] * v[0] + v[1] * v[1]);
            v[0] /= len;
            v[1] /= len;
        }
    }
}


